/**
 ** BigBlueButton open source conferencing system - http://www.bigbluebutton.org/
 *
 * Copyright (c) 2012 BigBlueButton Inc. and by respective authors (see below).
 *
 * This program is free software; you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation; either version 3.0 of the License, or (at your option) any later
 * version.
 *
 * BigBlueButton is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with BigBlueButton; if not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.bigbluebutton.web.controllers

import com.google.gson.Gson
import grails.web.context.ServletContextHolder
import groovy.json.JsonBuilder
import org.apache.commons.codec.binary.Base64
import org.apache.commons.io.FilenameUtils
import org.apache.commons.lang.RandomStringUtils
import org.apache.commons.lang.StringUtils
import org.bigbluebutton.api.*
import org.bigbluebutton.api.domain.Config
import org.bigbluebutton.api.domain.GuestPolicy
import org.bigbluebutton.api.domain.Meeting
import org.bigbluebutton.api.domain.UserSession
import org.bigbluebutton.api.util.ParamsUtil
import org.bigbluebutton.api.util.ResponseBuilder
import org.bigbluebutton.presentation.PresentationUrlDownloadService
import org.bigbluebutton.presentation.UploadedPresentation
import org.bigbluebutton.web.services.PresentationService
import org.bigbluebutton.web.services.turn.StunTurnService
import org.bigbluebutton.web.services.turn.TurnEntry
import org.bigbluebutton.web.services.turn.StunServer
import org.bigbluebutton.web.services.turn.RemoteIceCandidate
import org.json.JSONArray

import javax.servlet.ServletRequest

class ApiController {
  private static final Integer SESSION_TIMEOUT = 14400  // 4 hours
  private static final String CONTROLLER_NAME = 'ApiController'
  protected static final String RESP_CODE_SUCCESS = 'SUCCESS'
  protected static final String RESP_CODE_FAILED = 'FAILED'
  private static final String ROLE_MODERATOR = "MODERATOR"
  private static final String ROLE_ATTENDEE = "VIEWER"
  protected static Boolean REDIRECT_RESPONSE = true

  MeetingService meetingService;
  PresentationService presentationService
  ParamsProcessorUtil paramsProcessorUtil
  ClientConfigService configService
  PresentationUrlDownloadService presDownloadService
  StunTurnService stunTurnService
  HTML5LoadBalancingService html5LoadBalancingService
  ResponseBuilder responseBuilder = initResponseBuilder()

  def initResponseBuilder = {
    String protocol = this.getClass().getResource("").getProtocol();
    if (Objects.equals(protocol, "jar")) {
      // Application running inside a JAR file
      responseBuilder = new ResponseBuilder(getClass().getClassLoader(), "/WEB-INF/freemarker")
    } else if (Objects.equals(protocol, "file")) {
      // Application unzipped and running outside a JAR file
      String templateLoc = ServletContextHolder.servletContext.getRealPath("/WEB-INF/freemarker")
      // We should never have a null `templateLoc`
      responseBuilder = new ResponseBuilder(new File(templateLoc))
    }
  }

  /* general methods */
  def index = {
    log.debug CONTROLLER_NAME + "#index"
    response.addHeader("Cache-Control", "no-cache")

    withFormat {
      xml {
        render(text: responseBuilder.buildMeetingVersion(paramsProcessorUtil.getApiVersion(), RESP_CODE_SUCCESS), contentType: "text/xml")
      }
    }
  }

  /***********************************
   * CREATE (API)
   ***********************************/
  def create = {
    String API_CALL = 'create'
    log.debug CONTROLLER_NAME + "#${API_CALL}"
    log.debug request.getParameterMap().toMapString()

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors();
    paramsProcessorUtil.processRequiredCreateParams(params, errors);

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree with the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    // Ensure unique TelVoice. Uniqueness is not guaranteed by paramsProcessorUtil.
    if (!params.voiceBridge) {
      // Try up to 10 times. We should find a valid telVoice quickly unless
      // the server hosts ~100k meetings (default 5-digit telVoice)
      for (int i in 1..10) {
        String telVoice = paramsProcessorUtil.processTelVoice("");
        if (!meetingService.getNotEndedMeetingWithTelVoice(telVoice)) {
          params.voiceBridge = telVoice;
          break;
        }
      }
      // Still no unique voiceBridge found? Let createMeeting handle it.
    }

    params.html5InstanceId = html5LoadBalancingService.findSuitableHTML5ProcessByRoundRobin().toString()

    Meeting newMeeting = paramsProcessorUtil.processCreateParams(params)

    if (meetingService.createMeeting(newMeeting)) {
      // See if the request came with pre-uploading of presentation.
      uploadDocuments(newMeeting);  //
      respondWithConference(newMeeting, null, null)
    } else {
      // Translate the external meeting id into an internal meeting id.
      String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(params.meetingID);
      Meeting existing = meetingService.getNotEndedMeetingWithId(internalMeetingId);
      if (existing != null) {
        log.debug "Existing conference found"
        Map<String, Object> updateParams = paramsProcessorUtil.processUpdateCreateParams(params);
        if (existing.getViewerPassword().equals(params.get("attendeePW")) && existing.getModeratorPassword().equals(params.get("moderatorPW"))) {
          //paramsProcessorUtil.updateMeeting(updateParams, existing);
          // trying to create a conference a second time, return success, but give extra info
          // Ignore pre-uploaded presentations. We only allow uploading of presentation once.
          //uploadDocuments(existing);
          respondWithConference(existing, "duplicateWarning", "This conference was already in existence and may currently be in progress.");
        } else {
          // BEGIN - backward compatibility
          invalid("idNotUnique", "A meeting already exists with that meeting ID.  Please use a different meeting ID.");
          return
          // END - backward compatibility

          // enforce meetingID unique-ness
          errors.nonUniqueMeetingIdError()
          respondWithErrors(errors)
        }

        return
      } else {
        Meeting existingTelVoice = meetingService.getNotEndedMeetingWithTelVoice(newMeeting.getTelVoice());
        Meeting existingWebVoice = meetingService.getNotEndedMeetingWithWebVoice(newMeeting.getWebVoice());
        if (existingTelVoice != null || existingWebVoice != null) {
          log.error "VoiceBridge already in use by another meeting (different meetingId)"
          errors.nonUniqueVoiceBridgeError()
          respondWithErrors(errors)
        }
      }
    }
  }


  /**********************************************
   * JOIN API
   *********************************************/
  def join = {
    String API_CALL = 'join'
    log.debug CONTROLLER_NAME + "#${API_CALL}"
    ApiErrors errors = new ApiErrors()

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    //checking for an empty username or for a username containing whitespaces only
    if (!StringUtils.isEmpty(params.fullName)) {
      params.fullName = StringUtils.strip(params.fullName);
      if (StringUtils.isEmpty(params.fullName)) {
        invalid("missingParamFullName", "You must specify a name for the attendee who will be joining the meeting.", REDIRECT_RESPONSE);
        return
      }
    } else {
      invalid("missingParamFullName", "You must specify a name for the attendee who will be joining the meeting.", REDIRECT_RESPONSE);
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.", REDIRECT_RESPONSE)
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.", REDIRECT_RESPONSE)
      return
    }

    if (StringUtils.isEmpty(params.password)) {
      invalid("invalidPassword", "You either did not supply a password or the password supplied is neither the attendee or moderator password for this conference.", REDIRECT_RESPONSE);
      return
    }

    // END - backward compatibility

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    Boolean authenticated = false;

    Boolean guest = false;
    if (!StringUtils.isEmpty(params.guest)) {
      guest = Boolean.parseBoolean(params.guest)
    } else {
      // guest param has not been passed. Make user as
      // authenticated by default. (ralam july 3, 2018)
      authenticated = true
    }


    if (!StringUtils.isEmpty(params.auth)) {
      authenticated = Boolean.parseBoolean(params.auth)
    }

    // Do we have a name for the user joining? If none, complain.
    if (!StringUtils.isEmpty(params.fullName)) {
      if (StringUtils.isEmpty(params.fullName)) {
        errors.missingParamError("fullName");
      }
    } else {
      errors.missingParamError("fullName");
    }
    String fullName = ParamsUtil.stripHTMLTags(params.fullName)

    // Do we have a meeting id? If none, complain.
    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        errors.missingParamError("meetingID");
      }
    } else {
      errors.missingParamError("meetingID");
    }
    String externalMeetingId = params.meetingID

    // Do we have a password? If not, complain.
    String attPW = params.password
    if (StringUtils.isEmpty(attPW)) {
      errors.missingParamError("password");
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors, REDIRECT_RESPONSE)
      return
    }

    // Everything is good so far. Translate the external meeting id to an internal meeting id. If
    // we can't find the meeting, complain.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(externalMeetingId);

    log.info("Retrieving meeting ${internalMeetingId}")
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    if (meeting == null) {
      // BEGIN - backward compatibility
      invalid("invalidMeetingIdentifier", "The meeting ID that you supplied did not match any existing meetings", REDIRECT_RESPONSE);
      return
      // END - backward compatibility

      errors.invalidMeetingIdError();
      respondWithErrors(errors, REDIRECT_RESPONSE)
      return
    }

    // the createTime mismatch with meeting's createTime, complain
    // In the future, the createTime param will be required
    if (params.createTime != null) {
      long createTime = 0;
      try {
        createTime = Long.parseLong(params.createTime);
      } catch (Exception e) {
        log.warn("could not parse createTime param");
        createTime = -1;
      }
      if (createTime != meeting.getCreateTime()) {
        // BEGIN - backward compatibility
        invalid("mismatchCreateTimeParam", "The createTime parameter submitted mismatches with the current meeting.", REDIRECT_RESPONSE);
        return
        // END - backward compatibility

        errors.mismatchCreateTimeParam();
        respondWithErrors(errors, REDIRECT_RESPONSE);
        return
      }
    }

    // Is this user joining a meeting that has been ended. If so, complain.
    if (meeting.isForciblyEnded()) {
      // BEGIN - backward compatibility
      invalid("meetingForciblyEnded", "You can not re-join a meeting that has already been forcibly ended.  However, once the meeting is removed from memory (according to the timeout configured on this server, you will be able to once again create a meeting with the same meeting ID", REDIRECT_RESPONSE);
      return
      // END - backward compatibility

      errors.meetingForciblyEndedError();
      respondWithErrors(errors, REDIRECT_RESPONSE)
      return
    }

    // Now determine if this user is a moderator or a viewer.
    String role = null;
    if (meeting.getModeratorPassword().equals(attPW)) {
      role = Meeting.ROLE_MODERATOR
    } else if (meeting.getViewerPassword().equals(attPW)) {
      role = Meeting.ROLE_ATTENDEE
    }

    if (role == null) {
      // BEGIN - backward compatibility
      invalid("invalidPassword", "You either did not supply a password or the password supplied is neither the attendee or moderator password for this conference.", REDIRECT_RESPONSE);
      return
      // END - backward compatibility

      errors.invalidPasswordError()
      respondWithErrors(errors, REDIRECT_RESPONSE)
      return
    }

    // We preprend "w_" to our internal meeting Id to indicate that this is a web user.
    // For users joining using the phone, we will prepend "v_" so it will be easier
    // to distinguish users who doesn't have a web client. (ralam june 12, 2017)
    String internalUserID = "w_" + RandomStringUtils.randomAlphanumeric(12).toLowerCase()

    String authToken = RandomStringUtils.randomAlphanumeric(12).toLowerCase()

    String sessionToken = RandomStringUtils.randomAlphanumeric(16).toLowerCase()

    String externUserID = params.userID
    if (StringUtils.isEmpty(externUserID)) {
      externUserID = internalUserID
    }

    //Return a Map with the user custom data
    Map<String, String> userCustomData = paramsProcessorUtil.getUserCustomData(params);

    //Currently, it's associated with the externalUserID
    if (userCustomData.size() > 0)
      meetingService.addUserCustomData(meeting.getInternalId(), externUserID, userCustomData);

    String configxml = null;

    if (!StringUtils.isEmpty(params.configToken)) {
      Config conf = meeting.getConfig(params.configToken);
      if (conf == null) {
        // Check if this config is one of our pre-built config
        configxml = configService.getConfig(params.configToken)
        if (configxml == null) {
          // BEGIN - backward compatibility
          invalid("noConfigFound", "We could not find a config for this request.", REDIRECT_RESPONSE);
          return
          // END - backward compatibility

          errors.noConfigFound();
          respondWithErrors(errors);
        }
      } else {
        configxml = conf.config;
      }
    } else {
      Config conf = meeting.getDefaultConfig();
      configxml = conf.config;
    }

    // Do not fail if there's no default config.xml, needed for an HTML5 client only scenario

    String guestStatusVal = meeting.calcGuestStatus(role, guest, authenticated)

    UserSession us = new UserSession();
    us.authToken = authToken;
    us.internalUserId = internalUserID
    us.conferencename = meeting.getName()
    us.meetingID = meeting.getInternalId()
    us.externMeetingID = meeting.getExternalId()
    us.externUserID = externUserID
    us.fullname = fullName
    us.role = role
    us.conference = meeting.getInternalId()
    us.room = meeting.getInternalId()
    us.voicebridge = meeting.getTelVoice()
    us.webvoiceconf = meeting.getWebVoice()
    us.mode = "LIVE"
    us.record = meeting.isRecord()
    us.welcome = meeting.getWelcomeMessage()
    us.guest = guest
    us.authed = authenticated
    us.guestStatus = guestStatusVal
    us.logoutUrl = meeting.getLogoutUrl()
    us.configXML = configxml;

    if (!StringUtils.isEmpty(params.defaultLayout)) {
      us.defaultLayout = params.defaultLayout;
    }

    if (!StringUtils.isEmpty(params.avatarURL)) {
      us.avatarURL = params.avatarURL;
    } else {
      us.avatarURL = meeting.defaultAvatarURL
    }

    String meetingId = meeting.getInternalId()

    if (hasReachedMaxParticipants(meeting, us)) {
      // BEGIN - backward compatibility
      invalid("maxParticipantsReached", "The number of participants allowed for this meeting has been reached.", REDIRECT_RESPONSE);
      return
      // END - backward compatibility

      errors.maxParticipantsReached();
      respondWithErrors(errors, REDIRECT_RESPONSE);
      return;
    }

    // Register user into the meeting.
    meetingService.registerUser(
        us.meetingID,
        us.internalUserId,
        us.fullname,
        us.role,
        us.externUserID,
        us.authToken,
        us.avatarURL,
        us.guest,
        us.authed,
        guestStatusVal
    )

    //Identify which of these to logs should be used. sessionToken or user-token
    log.info("Session sessionToken for " + us.fullname + " [" + session[sessionToken] + "]")
    log.info("Session user-token for " + us.fullname + " [" + session['user-token'] + "]")
    session.setMaxInactiveInterval(SESSION_TIMEOUT);

    //check if exists the param redirect
    boolean redirectClient = true;
    String clientURL = paramsProcessorUtil.getDefaultHTML5ClientUrl();

    if (!StringUtils.isEmpty(params.redirect)) {
      try {
        redirectClient = Boolean.parseBoolean(params.redirect);
      } catch (Exception e) {
        redirectClient = true;
      }
    }

    String msgKey = "successfullyJoined"
    String msgValue = "You have joined successfully."

    // Keep track of the client url in case this needs to wait for
    // approval as guest. We need to be able to send the user to the
    // client after being approved by moderator.
    us.clientUrl = clientURL + "?sessionToken=" + sessionToken

    session[sessionToken] = sessionToken
    meetingService.addUserSession(sessionToken, us);

    // Process if we send the user directly to the client or
    // have it wait for approval.
    String destUrl = clientURL + "?sessionToken=" + sessionToken
    if (guestStatusVal.equals(GuestPolicy.WAIT)) {
      String guestWaitUrl = paramsProcessorUtil.getDefaultGuestWaitURL();
      destUrl = guestWaitUrl + "?sessionToken=" + sessionToken
      msgKey = "guestWait"
      msgValue = "Guest waiting for approval to join meeting."
    } else if (guestStatusVal.equals(GuestPolicy.DENY)) {
      destUrl = meeting.getLogoutUrl()
      msgKey = "guestDeny"
      msgValue = "Guest denied to join meeting."
    }

    Map<String, Object> logData = new HashMap<String, Object>();
    logData.put("meetingid", us.meetingID);
    logData.put("extMeetingid", us.externMeetingID);
    logData.put("name", us.fullname);
    logData.put("userid", us.internalUserId);
    logData.put("sessionToken", sessionToken);
    logData.put("logCode", "join_api");
    logData.put("description", "Handle JOIN API.");

    Gson gson = new Gson();
    String logStr = gson.toJson(logData);

    log.info(" --analytics-- data=" + logStr);

    if (redirectClient) {
      log.info("Redirecting to ${destUrl}");
      redirect(url: destUrl);
    } else {
      log.info("Successfully joined. Sending XML response.");
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildJoinMeeting(us, session[sessionToken], guestStatusVal, destUrl, msgKey, msgValue, RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    }
  }

  /*******************************************
   * IS_MEETING_RUNNING API
   *******************************************/
  def isMeetingRunning = {
    String API_CALL = 'isMeetingRunning'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a meeting id? If none, complain.
    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        errors.missingParamError("meetingID");
      }
    } else {
      errors.missingParamError("meetingID");
    }
    String externalMeetingId = params.meetingID


    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    // Everything is good so far. Translate the external meeting id to an internal meeting id. If
    // we can't find the meeting, complain.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(externalMeetingId);
    log.info("Retrieving meeting ${internalMeetingId}")
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    boolean isRunning = meeting != null && meeting.isRunning();

    response.addHeader("Cache-Control", "no-cache")
    withFormat {
      xml {
        render(contentType: "text/xml") {
          render(text: responseBuilder.buildIsMeetingRunning(isRunning, RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    }
  }

  /************************************
   * END API
   ************************************/
  def end = {
    String API_CALL = "end"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    if (StringUtils.isEmpty(params.password)) {
      invalid("invalidPassword", "You must supply the moderator password for this call.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a meeting id? If none, complain.
    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        errors.missingParamError("meetingID");
      }
    } else {
      errors.missingParamError("meetingID");
    }
    String externalMeetingId = params.meetingID

    // Do we have a password? If not, complain.
    String modPW = params.password
    if (StringUtils.isEmpty(modPW)) {
      errors.missingParamError("password");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    // Everything is good so far. Translate the external meeting id to an internal meeting id. If
    // we can't find the meeting, complain.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(externalMeetingId);
    log.info("Retrieving meeting ${internalMeetingId}")
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    if (meeting == null) {
      // BEGIN - backward compatibility
      invalid("notFound", "We could not find a meeting with that meeting ID - perhaps the meeting is not yet running?");
      return;
      // END - backward compatibility

      errors.invalidMeetingIdError();
      respondWithErrors(errors)
      return;
    }

    if (meeting.getModeratorPassword().equals(modPW) == false) {
      // BEGIN - backward compatibility
      invalid("invalidPassword", "You must supply the moderator password for this call.");
      return;
      // END - backward compatibility

      errors.invalidPasswordError();
      respondWithErrors(errors)
      return;
    }

    Map<String, Object> logData = new HashMap<String, Object>();
    logData.put("meetingid", meeting.getInternalId());
    logData.put("extMeetingid", meeting.getExternalId());
    logData.put("name", meeting.getName());
    logData.put("logCode", "end_api");
    logData.put("description", "Handle END API.");

    Gson gson = new Gson();
    String logStr = gson.toJson(logData);

    log.info(" --analytics-- data=" + logStr);

    meetingService.endMeeting(meeting.getInternalId());

    response.addHeader("Cache-Control", "no-cache")
    withFormat {
      xml {
        render(contentType: "text/xml") {
          render(text: responseBuilder.buildEndRunning("sentEndMeetingRequest", "A request to end the meeting was sent.  Please wait a few seconds, and then use the getMeetingInfo or isMeetingRunning API calls to verify that it was ended.", RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    }
  }

  /*****************************************
   * GETMEETINGINFO API
   *****************************************/
  def getMeetingInfo = {
    String API_CALL = "getMeetingInfo"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a meeting id? If none, complain.
    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        errors.missingParamError("meetingID");
      }
    } else {
      errors.missingParamError("meetingID");
    }
    String externalMeetingId = params.meetingID

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    // Everything is good so far. Translate the external meeting id to an internal meeting id. If
    // we can't find the meeting, complain.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(externalMeetingId);
    log.info("Retrieving meeting ${internalMeetingId}")
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    if (meeting == null) {
      // BEGIN - backward compatibility
      invalid("notFound", "We could not find a meeting with that meeting ID");
      return;
      // END - backward compatibility

      errors.invalidMeetingIdError();
      respondWithErrors(errors)
      return;
    }

    withFormat {
      xml {
        render(text: responseBuilder.buildGetMeetingInfoResponse(meeting, RESP_CODE_SUCCESS), contentType: "text/xml")
      }
    }
  }

  /************************************
   *  GETMEETINGS API
   ************************************/
  def getMeetingsHandler = {
    String API_CALL = "getMeetings"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    Collection<Meeting> mtgs = meetingService.getMeetings();

    if (mtgs == null || mtgs.isEmpty()) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildGetMeetingsResponse(mtgs, "noMeetings", "no meetings were found on this server", RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    } else {
      response.addHeader("Cache-Control", "no-cache")

      withFormat {
        xml {
          render(text: responseBuilder.buildGetMeetingsResponse(mtgs, null, null, RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    }
  }

  /************************************
   *  GETSESSIONS API
   ************************************/
  def getSessionsHandler = {
    String API_CALL = "getSessions"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    Collection<UserSession> sssns = meetingService.getSessions();

    if (sssns == null || sssns.isEmpty()) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildGetSessionsResponse(sssns, "noSessions", "no sessions were found on this serverr", RESP_CODE_SUCCESS), contentType: "text/xml")
        }
      }
    } else {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(contentType: "text/xml") {
            render(text: responseBuilder.buildGetSessionsResponse(sssns, null, null, RESP_CODE_SUCCESS), contentType: "text/xml")
          }
        }
      }
    }
  }


  private static Map<String, String[]> getParameters(ServletRequest request) {
    // Copy the parameters into our own Map as we can't pass the paramMap
    // from the request as it's an unmodifiable map.
    Map<String, String[]> reqParams = new HashMap<String, String[]>();
    Map<String, String[]> unModReqParams = request.getParameterMap();

    SortedSet<String> keys = new TreeSet<String>(unModReqParams.keySet());

    for(String key : keys) {
      reqParams.put(key, unModReqParams.get(key));
    }

    return reqParams;
  }

  /***********************************************
   * POLL API
   ***********************************************/
  def setPollXML = {
    String API_CALL = "setPollXML"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (StringUtils.isEmpty(params.pollXML)) {
      invalid("configXMLError", "You did not pass a poll XML")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    // Translate the external meeting id into an internal meeting id.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(params.meetingID);
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    if (meeting == null) {
      // BEGIN - backward compatibility
      invalid("invalidMeetingIdentifier", "The meeting ID that you supplied did not match any existing meetings");
      return;
      // END - backward compatibility
    }

    Map<String, String[]> reqParams = getParameters(request)

    String pollXML = params.pollXML

    String decodedPollXML;

    try {
      decodedPollXML = URLDecoder.decode(pollXML, "UTF-8");
    } catch (UnsupportedEncodingException e) {
      log.error "Couldn't decode poll XML.", e
      invalid("pollXMLError", "Cannot decode poll XML")
      return;
    }

    if (!paramsProcessorUtil.isPostChecksumSame(API_CALL, reqParams)) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          invalid("pollXMLChecksumError", "pollXMLChecksumError: request did not pass the checksum security check.")
        }
      }
    } else {
      def pollxml = new XmlSlurper().parseText(decodedPollXML);

      pollxml.children().each { poll ->
        String title = poll.title.text();
        String question = poll.question.text();
        String questionType = poll.questionType.text();

        ArrayList<String> answers = new ArrayList<String>();
        poll.answers.children().each { answer ->
          answers.add(answer.text());
        }

        //send poll to BigBlueButton Apps
        meetingService.createdPolls(meeting.getInternalId(), title, question, questionType, answers);
      }

      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          // No need to use the response builder here until we have a more complex response
          render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode></response>", contentType: "text/xml")
        }
      }
    }
  }

  /***********************************************
   * CONFIG API
   ***********************************************/
  def setConfigXML = {
    String API_CALL = "setConfigXML"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (StringUtils.isEmpty(params.configXML)) {
      invalid("configXMLError", "You did not pass a config XML")
      return
    }

    if (!StringUtils.isEmpty(params.meetingID)) {
      params.meetingID = StringUtils.strip(params.meetingID);
      if (StringUtils.isEmpty(params.meetingID)) {
        invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
        return
      }
    } else {
      invalid("missingParamMeetingID", "You must specify a meeting ID for the meeting.");
      return
    }

    // Translate the external meeting id into an internal meeting id.
    String internalMeetingId = paramsProcessorUtil.convertToInternalMeetingId(params.meetingID);
    Meeting meeting = meetingService.getMeeting(internalMeetingId);
    if (meeting == null) {
      // BEGIN - backward compatibility
      invalid("invalidMeetingIdentifier", "The meeting ID that you supplied did not match any existing meetings");
      return;
      // END - backward compatibility
    }

    Map<String, String[]> reqParams = getParameters(request)

    String configXML = params.configXML

    String decodedConfigXML;

    try {
      decodedConfigXML = URLDecoder.decode(configXML, "UTF-8");
    } catch (UnsupportedEncodingException e) {
      log.error("Couldn't decode config XML.");
      invalid("configXMLError", "Cannot decode config XML")
      return;
    }

    if (!paramsProcessorUtil.isPostChecksumSame(API_CALL, reqParams)) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          invalid("configXMLChecksumError", "configXMLChecksumError: request did not pass the checksum security check.")
        }
      }
    } else {
      boolean defaultConfig = false;

      if (!StringUtils.isEmpty(params.defaultConfig)) {
        try {
          defaultConfig = Boolean.parseBoolean(params.defaultConfig);
        } catch (Exception e) {
          defaultConfig = false;
        }
      }

      String token = meeting.storeConfig(defaultConfig, decodedConfigXML);
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          // No need to use the response builder here until we have a more complex response
          render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode><configToken>$token</configToken></response>", contentType: "text/xml")
        }
      }
    }
  }

  def getDefaultConfigXML = {

    String API_CALL = "getDefaultConfigXML"
    ApiErrors errors = new ApiErrors();

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility


    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    String defConfigXML = paramsProcessorUtil.getDefaultConfigXML();
    if (StringUtils.isEmpty(defConfigXML)) {
      // BEGIN - backward compatibility
      invalid("noConfigFound", "We could not find a config for this request.", REDIRECT_RESPONSE);
      return
      // END - backward compatibility

      errors.noConfigFound();
      respondWithErrors(errors);
    }

    response.addHeader("Cache-Control", "no-cache")
    render text: defConfigXML, contentType: 'text/xml'
  }

  def configXML = {
    String API_CALL = 'configXML'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    String logoutUrl = paramsProcessorUtil.getDefaultLogoutUrl()
    boolean reject = false
    String sessionToken = sanitizeSessionToken(params.sessionToken)
    if (!hasValidSession(sessionToken)) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildConfgXmlReject("Could not find conference.", logoutUrl, RESP_CODE_FAILED), contentType: "text/xml")
        }
      }
    } else {
      UserSession us = getUserSession(sessionToken)
      if (StringUtils.isEmpty(us.configXML)) {
        // BEGIN - backward compatibility
        invalid("noConfigFound", "We could not find a config for this request.", REDIRECT_RESPONSE);
        return
        // END - backward compatibility

        errors.noConfigFound();
        respondWithErrors(errors);
      }

      Map<String, Object> logData = new HashMap<String, Object>();
      logData.put("meetingId", us.meetingID);
      logData.put("externalMeetingId", us.externMeetingID);
      logData.put("name", us.fullname);
      logData.put("userId", us.internalUserId);
      logData.put("sessionToken", sessionToken);
      logData.put("logCode", "handle_configxml_api");
      logData.put("description", "Handling ConfigXml API.");

      Gson gson = new Gson();
      String logStr = gson.toJson(logData);

      log.info(" --analytics-- data=" + logStr);

      response.addHeader("Cache-Control", "no-cache")
      render text: us.configXML, contentType: 'text/xml'
    }
  }

  /**********************************************
   * GUEST WAIT API
   *********************************************/
  def guestWaitHandler = {
    String API_CALL = 'guestWait'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    ApiErrors errors = new ApiErrors()
    boolean reject = false;
    String sessionToken = sanitizeSessionToken(params.sessionToken)

    UserSession us = getUserSession(sessionToken);
    Meeting meeting = null;

    if (us == null) {
      log.debug("No user with session token.")
      reject = true;
    } else {
      meeting = meetingService.getMeeting(us.meetingID);
      if (meeting == null || meeting.isForciblyEnded()) {
        log.debug("Meeting not found.")
        reject = true
      }
    }

    // Determine the logout url so we can send the user there.
    String logoutUrl = us != null ? us.logoutUrl : paramsProcessorUtil.getDefaultLogoutUrl()

    if (reject) {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        json {
          def builder = new JsonBuilder()
          builder.response {
            returncode RESP_CODE_FAILED
            message "Could not process waiting guest."
            logoutURL logoutUrl
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    } else {
      //check if exists the param redirect
      boolean redirectClient = true;

      // Get the client url we stored in the join api call before
      // being told to wait.
      String clientURL = us.clientUrl;
      String lobbyMsg = meeting.getGuestLobbyMessage()
      log.info("clientURL = " + clientURL)
      log.info("redirect = ." + redirectClient)
      if (!StringUtils.isEmpty(params.redirect)) {
        try {
          redirectClient = Boolean.parseBoolean(params.redirect);
          log.info("redirect 2 = ." + redirectClient)
        } catch (Exception e) {
          redirectClient = true;
        }
      }

      // The client url is ovewriten. Let's allow it.
      if (!StringUtils.isEmpty(params.clientURL)) {
        clientURL = params.clientURL;
      }

      String guestWaitStatus = us.guestStatus

      log.debug("GuestWaitStatus = " + guestWaitStatus)

      String msgKey = "guestAllowed"
      String msgValue = "Guest allowed to join meeting."

      String destUrl = clientURL
      log.debug("destUrl = " + destUrl)

      if (guestWaitStatus.equals(GuestPolicy.WAIT)) {
        meetingService.guestIsWaiting(us.meetingID, us.internalUserId);
        clientURL = paramsProcessorUtil.getDefaultGuestWaitURL();
        destUrl = clientURL + "?sessionToken=" + sessionToken
        log.debug("GuestPolicy.WAIT - destUrl = " + destUrl)
        msgKey = "guestWait"
        msgValue = "Guest waiting for approval to join meeting."
        // We force the response to not do a redirect. Otherwise,
        // the client would just be redirecting into this endpoint.
        redirectClient = false

        Map<String, Object> logData = new HashMap<String, Object>();
        logData.put("meetingid", us.meetingID);
        logData.put("extMeetingid", us.externMeetingID);
        logData.put("name", us.fullname);
        logData.put("userid", us.internalUserId);
        logData.put("sessionToken", sessionToken);
        logData.put("logCode", "guest_wait");
        logData.put("description", "Guest waiting for approval.");

        Gson gson = new Gson();
        String logStr = gson.toJson(logData);

        log.info(" --analytics-- data=" + logStr);

      } else if (guestWaitStatus.equals(GuestPolicy.DENY)) {
        destUrl = meeting.getLogoutUrl()
        msgKey = "guestDenied"
        msgValue = "Guest denied to join meeting."

        Map<String, Object> logData = new HashMap<String, Object>();
        logData.put("meetingid", us.meetingID);
        logData.put("extMeetingid", us.externMeetingID);
        logData.put("name", us.fullname);
        logData.put("userid", us.internalUserId);
        logData.put("sessionToken", sessionToken);
        logData.put("logCode", "guest_denied");
        logData.put("description", "Guest denied.");

        Gson gson = new Gson();
        String logStr = gson.toJson(logData);

        log.info(" --analytics-- data=" + logStr);
      }

      if (redirectClient) {
        log.info("Redirecting to ${destUrl}");
        redirect(url: destUrl);
      } else {
        log.info("Successfully joined. Sending XML response.");
        response.addHeader("Cache-Control", "no-cache")
        withFormat {
          json {
            def builder = new JsonBuilder()
            builder.response {
              returncode RESP_CODE_SUCCESS
              messageKey msgKey
              message msgValue
              meeting_id us.meetingID
              user_id us.internalUserId
              auth_token us.authToken
              session_token session[sessionToken]
              guestStatus guestWaitStatus
              lobbyMessage lobbyMsg
              url destUrl
            }
            render(contentType: "application/json", text: builder.toPrettyString())
          }
        }
      }
    }
  }

  /***********************************************
   * ENTER API
   ***********************************************/
  def enter = {
    String API_CALL = 'enter'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    boolean reject = false;

    String sessionToken = sanitizeSessionToken(params.sessionToken)
    UserSession us = getUserSession(sessionToken);
    Meeting meeting = null;

    String respMessage = "Session not found."

    if (!hasValidSession(sessionToken)) {
      reject = true;
    } else {
      meeting = meetingService.getMeeting(us.meetingID);
      if (meeting == null || meeting.isForciblyEnded()) {
        reject = true
        respMessage = "Meeting not found or ended for session."
      } else {
        if (hasReachedMaxParticipants(meeting, us)) {
          reject = true;
          respMessage = "The number of participants allowed for this meeting has been reached.";
        } else {
          meeting.userEntered(us.internalUserId);
        }
      }
      if (us.guestStatus.equals(GuestPolicy.DENY)) {
        respMessage = "User denied for user with session."
        reject = true
      }
    }

    if (reject) {
      // Determine the logout url so we can send the user there.
      String logoutUrl = paramsProcessorUtil.getDefaultLogoutUrl()

      if (us != null) {
        logoutUrl = us.logoutUrl
      }

      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        json {
          def builder = new JsonBuilder()
          builder.response {
            returncode RESP_CODE_FAILED
            message respMessage
            sessionToken
            logoutURL logoutUrl
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    } else {

      Map<String, String> userCustomData = paramsProcessorUtil.getUserCustomData(params);

      // Generate a new userId for this user. This prevents old connections from
      // removing the user when the user reconnects after being disconnected. (ralam jan 22, 2015)
      // We use underscore (_) to associate userid with the user. We are also able to track
      // how many times a user reconnects or refresh the browser.
      String newInternalUserID = us.internalUserId //+ "_" + us.incrementConnectionNum()

      Map<String, Object> logData = new HashMap<String, Object>();
      logData.put("meetingid", us.meetingID);
      logData.put("extMeetingid", us.externMeetingID);
      logData.put("name", us.fullname);
      logData.put("userid", newInternalUserID);
      logData.put("sessionToken", sessionToken);
      logData.put("logCode", "handle_enter_api");
      logData.put("description", "Handling ENTER API.");

      Gson gson = new Gson();
      String logStr = gson.toJson(logData);

      log.info(" --analytics-- data=" + logStr);

      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        json {
          def builder = new JsonBuilder()
          builder.response {
            returncode RESP_CODE_SUCCESS
            fullname us.fullname
            confname us.conferencename
            meetingID us.meetingID
            externMeetingID us.externMeetingID
            externUserID us.externUserID
            internalUserID newInternalUserID
            authToken us.authToken
            role us.role
            guest us.guest
            guestStatus us.guestStatus
            conference us.conference
            room us.room
            voicebridge us.voicebridge
            dialnumber meeting.getDialNumber()
            webvoiceconf us.webvoiceconf
            mode us.mode
            record us.record
            isBreakout meeting.isBreakout()
            logoutTimer meeting.getLogoutTimer()
            allowStartStopRecording meeting.getAllowStartStopRecording()
            welcome us.welcome
            if (!StringUtils.isEmpty(meeting.moderatorOnlyMessage) && us.role.equals(ROLE_MODERATOR)) {
              modOnlyMessage meeting.moderatorOnlyMessage
            }
            if (!StringUtils.isEmpty(meeting.bannerText)) {
              bannerText meeting.getBannerText()
              bannerColor meeting.getBannerColor()
            }
            customLogoURL meeting.getCustomLogoURL()
            customCopyright meeting.getCustomCopyright()
            muteOnStart meeting.getMuteOnStart()
            allowModsToUnmuteUsers meeting.getAllowModsToUnmuteUsers()
            logoutUrl us.logoutUrl
            defaultLayout us.defaultLayout
            avatarURL us.avatarURL
            if (meeting.breakoutRoomsParams != null) {
              breakoutRooms {
                enabled meeting.breakoutRoomsParams.enabled
                record meeting.breakoutRoomsParams.record
                privateChatEnabled meeting.breakoutRoomsParams.privateChatEnabled
              }
            }
            customdata (
              meeting.getUserCustomData(us.externUserID).collect { k, v ->
                ["$k": v]
              }
            )
            metadata (
              meeting.getMetadata().collect { k, v ->
                ["$k": v]
              }
            )
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    }
  }

  /***********************************************
   * STUN/TURN API
   ***********************************************/
  def stuns = {
    String API_CALL = 'stuns'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    boolean reject = false;

    String sessionToken = sanitizeSessionToken(params.sessionToken)
    UserSession us = getUserSession(sessionToken);
    Meeting meeting = null;

    if (!hasValidSession(sessionToken)) {
      reject = true;
    } else {
      meeting = meetingService.getMeeting(us.meetingID);
      if (meeting == null || meeting.isForciblyEnded()) {
        reject = true
      }
    }

    if (reject) {
      String logoutUrl = paramsProcessorUtil.getDefaultLogoutUrl()

      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        json {
          def builder = new JsonBuilder()
          builder {
            returncode RESP_CODE_FAILED
            message "Could not find conference."
            logoutURL logoutUrl
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    } else {
      Set<StunServer> stuns = stunTurnService.getStunServers()
      Set<TurnEntry> turns = stunTurnService.getStunAndTurnServersFor(us.internalUserId)
      Set<RemoteIceCandidate> candidates = stunTurnService.getRemoteIceCandidates()

      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        json {
          def builder = new JsonBuilder()
          builder {
            stunServers (
              stuns.collect { stun ->
                [url: stun.url]
              }
            )
            turnServers (
              turns.collect { turn ->
                [
                  username: turn.username,
                  password: turn.password,
                  url: turn.url,
                  ttl: turn.ttl
                ]
              }
            )
            remoteIceCandidates (
              candidates.collect { candidate ->
                [ip: candidate.ip ]
              }
            )
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    }
  }

  /*************************************************
   * SIGNOUT API
   *************************************************/
  def signOut = {
    String API_CALL = 'signOut'
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    String sessionToken = sanitizeSessionToken(params.sessionToken)

    Meeting meeting = null;

    if (sessionToken != null) {

      UserSession us = meetingService.removeUserSessionWithAuthToken(sessionToken);
      if (us != null) {
        Map<String, Object> logData = new HashMap<String, Object>();
        logData.put("meetingid", us.meetingID);
        logData.put("extMeetingid", us.externMeetingID);
        logData.put("name", us.fullname);
        logData.put("userid", us.internalUserId);
        logData.put("sessionToken", sessionToken);
        logData.put("message", "handle_signout_api");
        logData.put("logCode", "signout_api");
        logData.put("description", "Handling SIGNOUT API.");

        Gson gson = new Gson();
        String logStr = gson.toJson(logData);
        log.info(" --analytics-- data=" + logStr);
      } else {
        log.info("Could not find user session for session token {}", sessionToken)
      }


      session.removeAttribute(sessionToken)
    }

    response.addHeader("Cache-Control", "no-cache")
    withFormat {
      xml {
        // No need to use the response builder here until we have a more complex response
        render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode></response>", contentType: "text/xml")
      }
    }
  }

  /******************************************************
   * GET_RECORDINGS API
   ******************************************************/
  def getRecordingsHandler = {
    String API_CALL = "getRecordings"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
      respondWithErrors(errors)
      return
    }

    log.debug request.getQueryString()

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    List<String> externalMeetingIds = new ArrayList<String>();
    if (!StringUtils.isEmpty(params.meetingID)) {
      externalMeetingIds = paramsProcessorUtil.decodeIds(params.meetingID);
    }

    ArrayList<String> internalRecordIds = new ArrayList<String>()
    if (!StringUtils.isEmpty(params.recordID)) {
      internalRecordIds = paramsProcessorUtil.decodeIds(params.recordID)
    }

    ArrayList<String> states = new ArrayList<String>()
    if (!StringUtils.isEmpty(params.state)) {
      states = paramsProcessorUtil.decodeIds(params.state)
    }

    // Everything is good so far.
    if (internalRecordIds.size() == 0 && externalMeetingIds.size() > 0) {
      // No recordIDs, process the request based on meetingID(s)
      // Translate the external meeting ids to internal meeting ids (which is the seed for the recordIDs).
      internalRecordIds = paramsProcessorUtil.convertToInternalMeetingId(externalMeetingIds);
    }

    for(String intRecId : internalRecordIds) {
      log.debug intRecId
    }

    Map<String, String> metadataFilters = ParamsProcessorUtil.processMetaParam(params);

    def getRecordingsResult = meetingService.getRecordings2x(internalRecordIds, states, metadataFilters)

    withFormat {
      xml {
        render(text: getRecordingsResult, contentType: "text/xml")
      }
    }
  }

  /******************************************************
   * PUBLISH_RECORDINGS API
   ******************************************************/
  def publishRecordings = {
    String API_CALL = "publishRecordings"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (StringUtils.isEmpty(params.recordID)) {
      invalid("missingParamRecordID", "You must specify a recordID.");
      return
    }

    if (StringUtils.isEmpty(params.publish)) {
      invalid("missingParamPublish", "You must specify a publish value true or false.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a recording id? If none, complain.
    String recordId = params.recordID
    if (StringUtils.isEmpty(recordId)) {
      errors.missingParamError("recordID");
    }
    // Do we have a publish status? If none, complain.
    String publish = params.publish
    if (StringUtils.isEmpty(publish)) {
      errors.missingParamError("publish");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    ArrayList<String> recordIdList = new ArrayList<String>();
    if (!StringUtils.isEmpty(recordId)) {
      recordIdList = paramsProcessorUtil.decodeIds(recordId);
    }

    if (!meetingService.existsAnyRecording(recordIdList)) {
      // BEGIN - backward compatibility
      invalid("notFound", "We could not find recordings");
      return;
      // END - backward compatibility

    }

    meetingService.setPublishRecording(recordIdList, publish.toBoolean());
    withFormat {
      xml {
        // No need to use the response builder here until we have a more complex response
        render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode><published>$publish</published></response>", contentType: "text/xml")
      }
    }
  }

  /******************************************************
   * DELETE_RECORDINGS API
   ******************************************************/
  def deleteRecordings = {
    String API_CALL = "deleteRecordings"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (StringUtils.isEmpty(params.recordID)) {
      invalid("missingParamRecordID", "You must specify a recordID.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a recording id? If none, complain.
    String recordId = params.recordID
    if (StringUtils.isEmpty(recordId)) {
      errors.missingParamError("recordID");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    List<String> recordIdList = new ArrayList<String>();
    if (!StringUtils.isEmpty(recordId)) {
      recordIdList = paramsProcessorUtil.decodeIds(recordId);
    }

    if (!meetingService.existsAnyRecording(recordIdList)) {
      // BEGIN - backward compatibility
      invalid("notFound", "We could not find recordings");
      return;
      // END - backward compatibility
    }

    meetingService.deleteRecordings(recordIdList);
    withFormat {
      xml {
        // No need to use the response builder here until we have a more complex response
        render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode><deleted>true</deleted></response>", contentType: "text/xml")
      }
    }
  }

  /******************************************************
   * UPDATE_RECORDINGS API
   ******************************************************/
  def updateRecordingsHandler = {
    String API_CALL = "updateRecordings"
    log.debug CONTROLLER_NAME + "#${API_CALL}"

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    // BEGIN - backward compatibility
    if (StringUtils.isEmpty(params.checksum)) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }

    if (StringUtils.isEmpty(params.recordID)) {
      invalid("missingParamRecordID", "You must specify a recordID.");
      return
    }

    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      invalid("checksumError", "You did not pass the checksum security check")
      return
    }
    // END - backward compatibility

    ApiErrors errors = new ApiErrors()

    // Do we have a checksum? If none, complain.
    if (StringUtils.isEmpty(params.checksum)) {
      errors.missingParamError("checksum");
    }

    // Do we have a recording id? If none, complain.
    String recordId = params.recordID
    if (StringUtils.isEmpty(recordId)) {
      errors.missingParamError("recordID");
    }

    if (errors.hasErrors()) {
      respondWithErrors(errors)
      return
    }

    // Do we agree on the checksum? If not, complain.
    if (!paramsProcessorUtil.isChecksumSame(API_CALL, params.checksum, request.getQueryString())) {
      errors.checksumError()
      respondWithErrors(errors)
      return
    }

    List<String> recordIdList = new ArrayList<String>();
    if (!StringUtils.isEmpty(recordId)) {
      recordIdList = paramsProcessorUtil.decodeIds(recordId);
    }

    if (!meetingService.existsAnyRecording(recordIdList)) {
      // BEGIN - backward compatibility
      invalid("notFound", "We could not find recordings");
      return;
      // END - backward compatibility
    }

    //Execute code specific for this call
    Map<String, String> metaParams = ParamsProcessorUtil.processMetaParam(params)
    if (!metaParams.empty) {
      //Proceed with the update
      meetingService.updateRecordings(recordIdList, metaParams);
    }
    withFormat {
      xml {
        // No need to use the response builder here until we have a more complex response
        render(text: "<response><returncode>$RESP_CODE_SUCCESS</returncode><updated>true</updated></response>", contentType: "text/xml")
      }
    }
  }

  def uploadDocuments(conf) { //
    log.debug("ApiController#uploadDocuments(${conf.getInternalId()})");

    //sanitizeInput
    params.each {
      key, value -> params[key] = sanitizeInput(value)
    }

    String requestBody = request.inputStream == null ? null : request.inputStream.text;
    requestBody = StringUtils.isEmpty(requestBody) ? null : requestBody;

    if (requestBody == null) {
      downloadAndProcessDocument(presentationService.defaultUploadedPresentation, conf.getInternalId(), true /* default presentation */, '');
    } else {
      def xml = new XmlSlurper().parseText(requestBody);
      xml.children().each { module ->
        log.debug("module config found: [${module.@name}]");

        if ("presentation".equals(module.@name.toString())) {
          // need to iterate over presentation files and process them
          Boolean current = true;
          module.children().each { document ->
            if (!StringUtils.isEmpty(document.@url.toString())) {
              def fileName;
              if (!StringUtils.isEmpty(document.@filename.toString())) {
                log.debug("user provided filename: [${module.@filename}]");
                fileName = document.@filename.toString();
              }
              downloadAndProcessDocument(document.@url.toString(), conf.getInternalId(), current /* default presentation */, fileName);
              current = false;
            } else if (!StringUtils.isEmpty(document.@name.toString())) {
              def b64 = new Base64()
              def decodedBytes = b64.decode(document.text().getBytes())
              processDocumentFromRawBytes(decodedBytes, document.@name.toString(),
                  conf.getInternalId(), current /* default presentation */);
              current = false;
            } else {
              log.debug("presentation module config found, but it did not contain url or name attributes");
            }
          }
        }
      }
    }
  }

  def processDocumentFromRawBytes(bytes, presOrigFilename, meetingId, current) {
    def uploadFailed = false
    def uploadFailReasons = new ArrayList<String>()

    // Gets the name minus the path from a full fileName.
    // a/b/c.txt --> c.txt
    def presFilename =  FilenameUtils.getName(presOrigFilename)
    def filenameExt = FilenameUtils.getExtension(presOrigFilename)
    def pres = null
    def presId = null

    if (presFilename == "" || filenameExt == "") {
      log.debug("Upload failed. Invalid filename " + presOrigFilename)
      uploadFailReasons.add("invalid_filename")
      uploadFailed = true
    } else {
      String presentationDir = presentationService.getPresentationDir()
      presId = Util.generatePresentationId(presFilename)

      File uploadDir = Util.createPresentationDir(meetingId, presentationDir, presId)
      if (uploadDir != null) {
        def newFilename = Util.createNewFilename(presId, filenameExt)
        pres = new File(uploadDir.absolutePath + File.separatorChar + newFilename);

        FileOutputStream fos = new java.io.FileOutputStream(pres)
        fos.write(bytes)
        fos.flush()
        fos.close()
      } else {
        log.warn "Upload failed. File Empty."
        uploadFailReasons.add("failed_to_download_file")
        uploadFailed = true
      }
    }

    // Hardcode pre-uploaded presentation to the default presentation window
    processUploadedFile("DEFAULT_PRESENTATION_POD",
              meetingId,
              presId,
              presFilename,
              pres,
              current,
              "preupload-raw-authz-token",
              uploadFailed,
              uploadFailReasons)
  }

  def downloadAndProcessDocument(address, meetingId, current, fileName) {
    log.debug("ApiController#downloadAndProcessDocument(${address}, ${meetingId}, ${fileName})");
    String presOrigFilename;
    if (StringUtils.isEmpty(fileName)) {
      presOrigFilename = address.tokenize("/")[-1];
    } else {
      presOrigFilename = fileName;
    }

    def uploadFailed = false
    def uploadFailReasons = new ArrayList<String>()

    // Gets the name minus the path from a full fileName.
    // a/b/c.txt --> c.txt
    def presFilename =  FilenameUtils.getName(presOrigFilename)
    def filenameExt = FilenameUtils.getExtension(presOrigFilename)
    def pres = null
    def presId

    if (presFilename == "" || filenameExt == "") {
      log.debug("Upload failed. Invalid filename " + presOrigFilename)
      uploadFailReasons.add("invalid_filename")
      uploadFailed = true
    } else {
      String presentationDir = presentationService.getPresentationDir()
      presId = Util.generatePresentationId(presFilename)
      File uploadDir = Util.createPresentationDir(meetingId, presentationDir, presId)
      if (uploadDir != null) {
        def newFilename = Util.createNewFilename(presId, filenameExt)
        def newFilePath = uploadDir.absolutePath + File.separatorChar + newFilename

        if (presDownloadService.savePresentation(meetingId, newFilePath, address)) {
          pres = new File(newFilePath)
        } else {
          log.error("Failed to download presentation=[${address}], meeting=[${meetingId}], fileName=[${fileName}]")
          uploadFailReasons.add("failed_to_download_file")
          uploadFailed = true
        }
      } else {
        log.error("Null presentation directory meeting=[${meetingId}], presentationDir=[${presentationDir}], presId=[${presId}]")
        uploadFailReasons.add("null_presentation_dir")
        uploadFailed = true
      }
    }

    // Hardcode pre-uploaded presentation to the default presentation window
    processUploadedFile(
            "DEFAULT_PRESENTATION_POD",
            meetingId,
            presId,
            presFilename,
            pres,
            current,
            "preupload-download-authz-token",
            uploadFailed,
            uploadFailReasons
    )
  }


  def processUploadedFile(podId, meetingId, presId, filename, presFile, current, authzToken, uploadFailed, uploadFailReasons ) {
    def presentationBaseUrl = presentationService.presentationBaseUrl
    // TODO add podId
    UploadedPresentation uploadedPres = new UploadedPresentation(podId,
            meetingId,
            presId,
            filename,
            presentationBaseUrl,
            current,
            authzToken,
            uploadFailed,
            uploadFailReasons)
    uploadedPres.setUploadedFile(presFile);
    presentationService.processUploadedPresentation(uploadedPres);
  }

  def beforeInterceptor = {
    if (paramsProcessorUtil.isServiceEnabled() == false) {
      log.info("apiNotEnabled: The API service and/or controller is not enabled on this server.  To use it, you must first enable it.")
      // TODO: this doesn't stop the request - so it generates invalid XML
      //      since the request continues and renders a second response
      invalid("apiNotEnabled", "The API service and/or controller is not enabled on this server.  To use it, you must first enable it.")
    }
  }

  def respondWithConference(meeting, msgKey, msg) {
    response.addHeader("Cache-Control", "no-cache")
    withFormat {
      xml {
        log.debug "Rendering as xml"
        render(text: responseBuilder.buildMeeting(meeting, msgKey, msg, RESP_CODE_SUCCESS), contentType: "text/xml")
      }
    }
  }

  def getUserSession(token) {
    if (token == null) {
      return null
    }

    UserSession us = meetingService.getUserSessionWithAuthToken(token)
    if (us == null) {
      log.info("Cannot find UserSession for token ${token}")
    }

    return us
  }

  private def sanitizeInput (input) {
    if(input == null)
      return

    if(!("java.lang.String".equals(input.getClass().getName())))
      return input

    StringUtils.strip(input.replaceAll("\\p{Cntrl}", ""));
  }

  def sanitizeSessionToken(param) {
    if (param == null) {
      log.info("sanitizeSessionToken: token is null")
      return null
    }

    if (StringUtils.isEmpty(param)) {
      log.info("sanitizeSessionToken: token is empty")
      return null
    }

    return StringUtils.strip(param)
  }

  private Boolean hasValidSession(token) {
    UserSession us = getUserSession(token)
    if (us == null) {
      return false
    }

    if (!session[token]) {
      log.info("Session for token ${token} not found")

      Boolean allowRequestsWithoutSession = paramsProcessorUtil.getAllowRequestsWithoutSession()
      if (!allowRequestsWithoutSession) {
        log.info("Meeting related to ${token} doesn't allow requests without session")
        return false
      }
    }

    log.info("Token ${token} is valid")
    return true
  }

  // Validate maxParticipants constraint
  private Boolean hasReachedMaxParticipants(meeting, us) {
    // Meeting object calls it maxUsers to build up the drama
    int maxParticipants = meeting.getMaxUsers();
    // When is set to 0, the validation is ignored
    Boolean enabled = maxParticipants > 0;
    // Users refreshing page or reconnecting must be identified
    Boolean rejoin = meeting.getUserById(us.internalUserId) != null;
    // Users that passed enter once, still not joined but somehow re-entered
    Boolean reenter = meeting.getEnteredUserById(us.internalUserId) != null;
    // Users that already joined the meeting
    int joinedUsers = meeting.getUsers().size()
    // Users that are entering the meeting
    int enteredUsers = meeting.getEnteredUsers().size()

    Boolean reachedMax = (joinedUsers + enteredUsers) >= maxParticipants;
    if (enabled && !rejoin && !reenter && reachedMax) {
      return true;
    }

    return false;
  }

  private void respondWithErrors(errorList, redirectResponse = false) {
    log.debug CONTROLLER_NAME + "#invalid"
    if (redirectResponse) {
      ArrayList<Object> errors = new ArrayList<Object>();
      errorList.getErrors().each { error ->
        Map<String, String> errorMap = new LinkedHashMap<String, String>()
        errorMap.put("key", error[0])
        errorMap.put("message", error[1])
        errors.add(errorMap)
      }

      JSONArray errorsJSONArray = new JSONArray(errors);
      log.debug errorsJSONArray

      respondWithRedirect(errorsJSONArray)
    } else {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildErrors(errorList.getErrors(), RESP_CODE_FAILED), contentType: "text/xml")
        }
        json {
          log.debug "Rendering as json"
          def builder = new JsonBuilder()
          builder.response {
            returncode RESP_CODE_FAILED
            messageKey key
            message msg
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    }
  }

  //TODO: method added for backward compatibility, it will be removed in next versions after 0.8
  private void invalid(key, msg, redirectResponse = false) {
    // Note: This xml scheme will be DEPRECATED.
    log.debug CONTROLLER_NAME + "#invalid " + msg
    if (redirectResponse) {
      ArrayList<Object> errors = new ArrayList<Object>();
      Map<String, String> errorMap = new LinkedHashMap<String, String>()
      errorMap.put("key", key)
      errorMap.put("message", msg)
      errors.add(errorMap)

      JSONArray errorsJSONArray = new JSONArray(errors)
      log.debug "JSON Errors {}", errorsJSONArray.toString()

      respondWithRedirect(errorsJSONArray)
    } else {
      response.addHeader("Cache-Control", "no-cache")
      withFormat {
        xml {
          render(text: responseBuilder.buildError(key, msg, RESP_CODE_FAILED), contentType: "text/xml")
        }
        json {
          log.debug "Rendering as json"
          def builder = new JsonBuilder()
          builder.response {
            returncode RESP_CODE_FAILED
            messageKey key
            message msg
          }
          render(contentType: "application/json", text: builder.toPrettyString())
        }
      }
    }
  }

  private void respondWithRedirect(errorsJSONArray) {
    String logoutUrl = paramsProcessorUtil.getDefaultLogoutUrl()
    URI oldUri = URI.create(logoutUrl)

    if (!StringUtils.isEmpty(params.logoutURL)) {
      try {
        oldUri = URI.create(params.logoutURL)
      } catch (Exception e) {
        // Do nothing, the variable oldUri was already initialized
      }
    }

    String newQuery = oldUri.getQuery();

    if (newQuery == null) {
      newQuery = "errors="
    } else {
      newQuery += "&" + "errors="
    }
    newQuery += errorsJSONArray

    URI newUri = new URI(oldUri.getScheme(), oldUri.getAuthority(), oldUri.getPath(), newQuery, oldUri.getFragment())

    log.debug "Constructed logout URL {}", newUri.toString()
    redirect(url: newUri)
  }

}
